# Copyright 2021 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import copy

from model_compression_toolkit import CoreConfig
from model_compression_toolkit.core.common import Graph, BaseNode
from model_compression_toolkit.core.common.framework_implementation import FrameworkImplementation


def apply_bias_correction_to_graph(graph_to_apply_bias_correction: Graph,
                                   core_config: CoreConfig,
                                   fw_impl: FrameworkImplementation) -> Graph:
    """
    Get a graph, where each node has a final weights quantization configuration (with a bias
    correction term in it), and apply the bias correction for each node in the graph.

    Args:
        graph_to_apply_bias_correction: Graph to apply bias correction to.
        core_config: CoreConfig containing parameters of how the model should be quantized.
        fw_impl: FrameworkImplementation object with a specific framework methods implementation.

    Returns:
        Graph with bias correction apply to it's nodes.
    """

    graph = copy.deepcopy(graph_to_apply_bias_correction)
    for n in graph.nodes:
        if n.is_weights_quantization_enabled() and core_config.quantization_config.weights_bias_correction \
                and not n.final_weights_quantization_cfg.weights_second_moment_correction:
            # If a kernel was quantized and weights bias correction is enabled in n.quantization_cfg,
            # a bias correction term was calculated during model preparation, and is used now in the node's bias term.
            if n.final_weights_quantization_cfg.weights_bias_correction:
                _apply_bias_correction_to_node(n, fw_impl)
    return graph


def _apply_bias_correction_to_node(node:BaseNode,
                                   fw_impl: FrameworkImplementation):
    """
    Set new bias to node using the bias-correction term that is stored in the
    final weights quantization configuration.

    Args:
        node: Node to set its corrected bias after bias-correction.
        fw_impl: FrameworkImplementation object with a specific framework methods implementation.

    """
    correction = node.final_weights_quantization_cfg.bias_corrected

    bias = node.get_weights_by_keys(fw_impl.constants.BIAS)  # get original bias from node's weights

    if bias is not None:  # It the layer has bias, we subtract the correction from original bias
        node.set_weights_by_keys(fw_impl.constants.BIAS, node.get_weights_by_keys(fw_impl.constants.BIAS) - correction)

    else:  # It the layer has no bias, we consider it as if it has and its value is 0.
        node.set_weights_by_keys(fw_impl.constants.BIAS, - correction)
        node.framework_attr[fw_impl.constants.USE_BIAS] = True  # Mark the use_bias attribute of the node.



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Dict, Any

from torch.nn import BatchNorm2d, Conv2d, ConvTranspose2d

from model_compression_toolkit.core.common import BaseNode
from model_compression_toolkit.core.common.graph.graph_matchers import NodeOperationMatcher
from model_compression_toolkit.core.common.substitutions.batchnorm_reconstruction import \
    BatchNormalizationReconstruction
from model_compression_toolkit.core.pytorch.constants import GAMMA, BETA, MOVING_MEAN, MOVING_VARIANCE, \
    EPSILON, NUM_FEATURES, OUT_CHANNELS, EPSILON_VAL, MOMENTUM, MOMENTUM_VAL


def batchnorm_reconstruction_node_matchers():
    """
    Match linear layers for BN reconstruction.

    Returns:
        Node matcher for convolution type nodes.
    """
    source_node = NodeOperationMatcher(Conv2d) | \
                  NodeOperationMatcher(ConvTranspose2d)
    return source_node


def create_bn_node(source_node: BaseNode,
                   bn_node_weights: Dict[Any, Any]):
    """
    Creates BN nodes after linear nodes.
    Args:
        source_node: Node matcher for convolution type nodes.
        bn_node_weights: dictionary of BN weights.
    """
    bn_node = BaseNode(name=source_node.name + '_reconstructed',
                       framework_attr={NUM_FEATURES: source_node.framework_attr[OUT_CHANNELS],
                                       EPSILON: EPSILON_VAL,
                                       MOMENTUM: MOMENTUM_VAL},
                       input_shape=source_node.output_shape,
                       output_shape=source_node.output_shape,
                       weights=bn_node_weights,
                       layer_class=BatchNorm2d)
    return bn_node


def pytorch_batchnorm_reconstruction() -> BatchNormalizationReconstruction:
    """
    Reconstruct BatchNormalization after linear layers.
    Returns:
        A BatchNormalizationReconstruction initialized for Pytorch models.
    """
    source_node = batchnorm_reconstruction_node_matchers()
    return BatchNormalizationReconstruction(source_node, create_bn_node, GAMMA, BETA,
                                            MOVING_MEAN, MOVING_VARIANCE, EPSILON_VAL)



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Union, List
from abc import abstractmethod
import torch
import numpy as np
from model_compression_toolkit.core.common import Logger
from model_compression_toolkit.gptq.common.gptq_config import GradientPTQConfig
from model_compression_toolkit.gptq.pytorch.quantizer.gptq_quantizer import BaseWeightQuantizer
from model_compression_toolkit.core.common.quantization.node_quantization_config import NodeWeightsQuantizationConfig
from model_compression_toolkit.gptq.pytorch.quantizer.quant_utils import sample_gumbel
from model_compression_toolkit.core.pytorch.utils import to_torch_tensor
from model_compression_toolkit.core.common.target_platform.op_quantization_config import QuantizationMethod

P_INIT = 0.01

def init_aux_var(w_shape: torch.Size, m: int, p: float = P_INIT) -> torch.Tensor:
    """
    This function generate a random pi matrix for Gumbel Rounding
    Args:
        w_shape(torch.Size): A list of integers that represent the shape of the weights tensor to be quantization
        p(float): A floating point number that represent the probability of non-round options of pi matrix.
        m(int):  An integer that define the number of shift

    Returns: A torch tensor of pi tensor

    """
    m_hat = m // 2
    shift = -np.log(-np.log(1 - p))
    n = torch.randn([m, *w_shape]) * np.sqrt(np.square(torch.pi) / 6)
    n[m_hat, ...] += shift
    return n


def init_shift_var(m: int) -> torch.Tensor:
    """
    This function generate a tensor of 2*m+1 from -m to m
    Args:
        m: An integer value the represent m

    Returns: A tensor of size m

    """
    m_hat = m // 2
    aux_index_shift = [-m_hat + i for i in range(m)]
    return torch.Tensor(aux_index_shift)


class BaseGumbelWeightQuantizer(BaseWeightQuantizer):
    """
    Base class that implements a quantizer with trainable parameters to be used for GPTQ training.
    """

    def __init__(self,
                 weights_quantization_cfg: NodeWeightsQuantizationConfig,
                 gptq_config: GradientPTQConfig,
                 weight_shape: torch.Size):
        """
        Construct a Pytorch model that utilize a fake weight quantizer of Gumbel rounding
        Args:
            weights_quantization_cfg: Configuration of weight quantization
            gptq_config: GradientPTQConfig object with parameters about the tuning process.
            weight_shape: weight shape for auxiliary tensor creation.
        """
        super().__init__()

        self.power_of_two = QuantizationMethod.POWER_OF_TWO == weights_quantization_cfg.weights_quantization_method
        self.reshape_aux_shift = [-1, *[1 for _ in range(len(weight_shape))]]
        self.num_bits = weights_quantization_cfg.weights_n_bits
        self.weight_shape = weight_shape
        self.max_delta_change = gptq_config.lsb_change_per_bit_width.get(self.num_bits)
        self.quantization_parameter_learning = gptq_config.quantization_parameters_learning
        self.m = 2 * self.max_delta_change + 1
        self.minimal_temp = gptq_config.quantizer_config.minimal_temp
        self.maximal_temp = gptq_config.quantizer_config.maximal_temp
        self.temperature_learning = gptq_config.quantizer_config.temperature_learning
        self.cycle_iterations = int(gptq_config.n_iter / gptq_config.quantizer_config.n_cycles)
        self.shift_tensor = to_torch_tensor(init_shift_var(self.m))
        self.tau = None
        self.g_t = 0
        self.p_t = None
        self.n_iter = 0
        self.update_gumbel_param = True
        scale = self.cycle_iterations / (-2 * np.log(0.001))

        def tau_function(i: int) -> float:
            """
            A function that generates the gumbel temperature.
            Args:
                i: An int that represents the current iteration number

            Returns: A temperature value.

            """
            if i < (self.cycle_iterations - 1):
                index = ((i + 1) % self.cycle_iterations) / scale
            else:
                index = (i % self.cycle_iterations) / scale

            x = np.exp(-index)
            return self.minimal_temp + (self.maximal_temp - self.minimal_temp) * x

        self.tau_function = tau_function

    def get_gumbel_probability(self) -> torch.Tensor:
        """
        A function that return the gumbel probability value.
        Returns: gumbel probability
        """
        return self.p_t

    def update_iteration(self, training):
        """
        A function that update parameters for GPTQ fine-tuning
        Args:
            training: whether in training mode or not
        """
        if self.temperature_learning:
            self.tau = torch.clip(self.temp_tensor, self.minimal_temp, self.maximal_temp)
        else:
            self.tau = self.tau_function(self.n_iter)
        if self.update_gumbel_param and training:
            if self.cycle_iterations > 0 and self.n_iter % self.cycle_iterations == 0:
                self.temp_tensor.data = self.maximal_temp * to_torch_tensor(torch.ones(self.temp_tensor.shape))
            self.n_iter += 1
            self.g_t = sample_gumbel([self.m, *self.weight_shape])

    @abstractmethod
    def get_temperature_variable(self) -> Union[torch.Tensor, List]:
        """
        Returns temperature trainable variables
        """
        raise Logger.error(f"{self.__class__.__name__} have to implement this abstract function.")


# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import numpy as np
import torch

from model_compression_toolkit.core.common.quantization.quantizers.quantizers_helpers import uniform_quantize_tensor
from model_compression_toolkit.core.common.target_platform import QuantizationMethod


class UniformWeightsQuantizer:
    """
    Uniform quantizer for a PyTorch weights module.
    """
    def __init__(self,
                 num_bits: int,
                 min_range: np.ndarray,
                 max_range: np.ndarray,
                 quantization_method: QuantizationMethod,
                 per_channel: bool,
                 output_channels_axis: int
                 ):
        """

        Args:
            num_bits: Number of bits for quantization.
            min_range: Min range for quantization.
            max_range: Max range for quantization.
            quantization_method: Method the quantized weight was quantized according to.
            per_channel: Whether the weight was quantized per-channel or not.
            output_channels_axis: Dimension of channel axis (needed if it was quantized per-channel).
        """

        super(UniformWeightsQuantizer, self).__init__()
        self.num_bits = num_bits
        self.min_range = min_range
        self.max_range = max_range
        self.quantization_method = quantization_method
        self.per_channel = per_channel
        self.output_channels_axis = output_channels_axis

    def __call__(self, float_weight: np.ndarray, *args, **kwargs) -> np.ndarray:
        """
        Quantize tensor according to quantization params in the quantizer.

        Args:
            float_weight: Weights to quantize.

        Returns:
            Quantized tensor.

        """
        return uniform_quantize_tensor(tensor_data=float_weight,
                                       range_min=self.min_range,
                                       range_max=self.max_range,
                                       n_bits=self.num_bits)



# Copyright 2021 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from tensorflow_model_optimization.python.core.quantization.keras.quantizers import Quantizer
from typing import List, Any, Dict
from tensorflow import Tensor
import six, abc


@six.add_metaclass(abc.ABCMeta)
class BaseTrainableQuantizer(Quantizer):
    """
    Base trainable quantizer to define extra methods needed by the GPTQ post-processing.
    """

    @abc.abstractmethod
    def get_quant_config(self, layer) -> Dict[str, Any]:
        """
        Returns the config used to edit NodeQuantizationConfig after GPTQ retraining

        Args:
            layer: quantized layer

        Returns:
            A dictionary of attributes the quantize_config retraining has changed during GPTQ retraining.
            Keys must match NodeQuantizationConfig attributes

        """

    @abc.abstractmethod
    def get_trainable_parameters(self) -> List[Tensor]:
        """
        A function to get a list trainable of trainable parameters for GPTQ retraining

        Returns:
            A list of trainable Tensors

        """




# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import List, Callable

from model_compression_toolkit.core.pytorch.back2framework.quantization_wrapper.wrapper_quantize_config import \
    WrapperQuantizeConfig


class WeightsQuantizeConfig(WrapperQuantizeConfig):
    """
    QuantizationConfig for Fully Quantized model to quantize layer's weights only.
    """

    def __init__(self,
                 weight_quantizers: List[Callable]):
        """

        Args:
            weight_quantizers: List of quantizers to quantize the layer's weights.
        """
        super().__init__(is_weight_quantized=True,
                         is_activation_quantized=False)

        self._weight_quantizers = weight_quantizers

    def get_weight_quantizers(self) -> List[Callable]:
        """

        Returns: List of quantizers to quantize the layer's weights.

        """
        return self._weight_quantizers

    def get_activation_quantizer(self):
        """

        Returns: An empty list as this QC does not quantize the layer's activations.

        """
        return []



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import tensorflow as tf
from packaging import version

if version.parse(tf.__version__) < version.parse("2.6"):
    from tensorflow.keras.layers import Conv2D, DepthwiseConv2D, Reshape, ZeroPadding2D, \
        Dropout, \
        MaxPooling2D, Activation, ReLU, Flatten, Cropping2D
else:
    from keras.layers import Conv2D, DepthwiseConv2D, Reshape, ZeroPadding2D, \
        Dropout, MaxPooling2D, Activation, ReLU, Flatten, Cropping2D

from model_compression_toolkit.core.tpc_models.default_tpc.v1.tp_model import get_tp_model
import model_compression_toolkit as mct
from model_compression_toolkit.core.tpc_models.default_tpc.v1 import __version__ as TPC_VERSION

tp = mct.target_platform


def get_keras_tpc() -> tp.TargetPlatformCapabilities:
    """
    get a Keras TargetPlatformCapabilities object with default operation sets to layers mapping.
    Returns: a Keras TargetPlatformCapabilities object for the given TargetPlatformModel.
    """
    default_tp_model = get_tp_model()
    return generate_keras_tpc(name='default_keras_tpc', tp_model=default_tp_model)


def generate_keras_tpc(name: str, tp_model: tp.TargetPlatformModel):
    """
    Generates a TargetPlatformCapabilities object with default operation sets to layers mapping.

    Args:
        name: Name of the TargetPlatformCapabilities.
        tp_model: TargetPlatformModel object.

    Returns: a TargetPlatformCapabilities object for the given TargetPlatformModel.
    """

    keras_tpc = tp.TargetPlatformCapabilities(tp_model,
                                              name=name,
                                              version=TPC_VERSION)

    with keras_tpc:
        tp.OperationsSetToLayers("NoQuantization", [Reshape,
                                                    tf.reshape,
                                                    Flatten,
                                                    Cropping2D,
                                                    ZeroPadding2D,
                                                    Dropout,
                                                    MaxPooling2D,
                                                    tf.split,
                                                    tf.quantization.fake_quant_with_min_max_vars,
                                                    tf.math.argmax,
                                                    tf.shape,
                                                    tf.__operators__.getitem,
                                                    tf.compat.v1.shape])

        tp.OperationsSetToLayers("Conv", [Conv2D,
                                          DepthwiseConv2D,
                                          tf.nn.conv2d,
                                          tf.nn.depthwise_conv2d])
        tp.OperationsSetToLayers("AnyReLU", [tf.nn.relu,
                                             tf.nn.relu6,
                                             tp.LayerFilterParams(ReLU, negative_slope=0.0),
                                             tp.LayerFilterParams(Activation, activation="relu")])

    return keras_tpc



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import copy
from typing import Callable, Any

from tqdm import tqdm

from model_compression_toolkit.core import common
from model_compression_toolkit.core.common import FrameworkInfo
from model_compression_toolkit.core.common import Graph
from model_compression_toolkit.core.common.framework_implementation import FrameworkImplementation
from model_compression_toolkit.core.common.model_builder_mode import ModelBuilderMode
from model_compression_toolkit.core.common.model_collector import ModelCollector
from model_compression_toolkit.core.common.quantization.core_config import CoreConfig
from model_compression_toolkit.core.common.quantization.quantization_analyzer import analyzer_graph
from model_compression_toolkit.core.common.quantization.quantization_params_generation.qparams_activations_computation \
    import get_activations_qparams
from model_compression_toolkit.core.common.quantization.quantize_graph_weights import quantize_graph_weights
from model_compression_toolkit.core.common.substitutions.apply_substitutions import substitute


def _collect_and_assign_act_threshold(graph: Graph,
                                      representative_data_gen: Callable,
                                      core_config: CoreConfig,
                                      fw_info: FrameworkInfo,
                                      fw_impl: FrameworkImplementation):
    """
    Collect statistics after second moment correction and assign new thresholds to activations.
     Args:
        graph: Graph to apply second moment correction.
        representative_data_gen (Callable): Dataset used for calibration.
        core_config (CoreConfig): Configuration object containing parameters of how the model should be
         quantized, including mixed precision parameters.
        fw_info: FrameworkInfo object with information about the specific framework's model.
        fw_impl: FrameworkImplementation object with a specific framework methods implementation.
     """

    analyzer_graph(fw_impl.attach_sc_to_node,
                   graph,
                   fw_info,
                   core_config.quantization_config)  # Mark points for statistics collection

    mi = ModelCollector(graph,
                        fw_impl,
                        fw_info)

    for _ in tqdm(range(core_config.n_iter)):
        mi.infer(representative_data_gen())

    for n in list(graph.nodes):
        if n.is_activation_quantization_enabled():
            activation_params = get_activations_qparams(
                activation_quant_cfg=n.final_activation_quantization_cfg,
                nodes_prior_info=n.prior_info,
                out_stats_container=graph.get_out_stats_collector(n))
            n.final_activation_quantization_cfg.set_activation_quantization_param(activation_params)


def quantized_model_builder_for_second_moment_correction(graph: common.Graph,
                                                         fw_info: FrameworkInfo,
                                                         fw_impl: Any):
    """
    Build a framework model from a graph for second moment correction.

    Args:
        graph: Graph to build the from.
        fw_info: FrameworkInfo object with information about the specific framework's model.
        fw_impl: FrameworkImplementation object with a specific framework methods implementation.

    Returns:
        Quantized model for second moment correction.
    """
    quantized_tg = quantize_graph_weights(graph,
                                          fw_info=fw_info,
                                          fw_impl=fw_impl)

    quantized_model, user_info = fw_impl.model_builder(quantized_tg,
                                                       mode=ModelBuilderMode.FLOAT,
                                                       fw_info=fw_info)
    return quantized_model


def apply_second_moment_correction_to_graph(graph_to_apply_second_moment_correction: Graph,
                                            representative_data_gen: Callable,
                                            core_config: CoreConfig,
                                            fw_info: FrameworkInfo,
                                            fw_impl: FrameworkImplementation) -> Graph:
    """
     Apply second moment correction on graph.
     Args:
        graph_to_apply_second_moment_correction: Graph to apply second moment correction.
        representative_data_gen (Callable): Dataset used for calibration.
        core_config (CoreConfig): Configuration object containing parameters of how the model should be
         quantized, including mixed precision parameters.
        fw_info: FrameworkInfo object with information about the specific framework's model.
        fw_impl: FrameworkImplementation object with a specific framework methods implementation.

     Returns:
         Graph after second moment correction.
     """
    graph = copy.deepcopy(graph_to_apply_second_moment_correction)
    semi_quantized_model = quantized_model_builder_for_second_moment_correction(graph, fw_info, fw_impl)
    fw_impl.apply_second_moment_correction(semi_quantized_model, core_config, representative_data_gen, graph)
    graph = substitute(graph, fw_impl.get_substitutions_after_second_moment_correction(core_config.quantization_config))
    _collect_and_assign_act_threshold(graph, representative_data_gen, core_config, fw_info, fw_impl)

    return graph



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import torch.nn
from torch.nn import Conv2d, Linear, ConvTranspose2d
from model_compression_toolkit.core.common.graph.graph_matchers import NodeOperationMatcher
from model_compression_toolkit.core.common.substitutions.weights_activation_split import BaseWeightsActivationSplit

"""
Matches: (DepthwiseConv2D, Conv2D, Dense, Conv2DTranspose)
"""
OP2D_NODE = NodeOperationMatcher(Conv2d) | \
            NodeOperationMatcher(Linear) | \
            NodeOperationMatcher(ConvTranspose2d)


class WeightsActivationSplit(BaseWeightsActivationSplit):
    """
    Substitution extends BaseWeightsActivationSplit for Keras framework.
    """

    def __init__(self):
        """
        Initializes a WeightsActivationSplit substitution for Keras framework.
        """

        super().__init__(activation_layer_type=torch.nn.Identity,
                         fw_attr={},
                         matcher_instance=OP2D_NODE)



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from typing import List, Any, Tuple

import torch

from model_compression_toolkit.core import common
from model_compression_toolkit.core.common import BaseNode, Graph
from model_compression_toolkit.core.common.user_info import UserInformation
from model_compression_toolkit.core.pytorch.back2framework.instance_builder import node_builder
from model_compression_toolkit.core.pytorch.back2framework.pytorch_model_builder import PyTorchModelBuilder, \
    PytorchModel
from model_compression_toolkit.core.pytorch.back2framework.quantization_wrapper.quantized_layer_wrapper import \
    QuantizedLayerWrapper
from model_compression_toolkit.core.pytorch.constants import BUFFER, CONSTANT
from model_compression_toolkit.core.pytorch.reader.node_holders import BufferHolder, ConstantHolder
from model_compression_toolkit.core.pytorch.utils import get_working_device


from model_compression_toolkit.exporter.fully_quantized.pytorch.builder.node_to_quantize_config import get_quantization_config


def get_fully_quantized_pytorch_model(graph: Graph):
    """
    Convert graph to fully quantized PyTorch model.

    Args:
        graph: Graph to convert to a PyTorch model.

    Returns:
        Fully quantized PyTorch model.
    """
    return FullyQuantizedPyTorchModelBuilder(graph=graph).build_model()



class FullyQuantizedPyTorchModel(PytorchModel):
    """
    PyTorch model with all quantization information.
    """

    def __init__(self,
                 graph: common.Graph):
        """

        Args:
            graph: Graph to build its corresponding Pytorch model.
        """

        super().__init__(graph)


    def _add_modules(self):
        """

        Add nodes in graph as modules.

        """
        for n in self.node_sort:
            if n.type == BufferHolder:
                self.add_module(n.name, node_builder(n))
                self.get_submodule(n.name).register_buffer(n.name,
                                                           torch.Tensor(n.get_weights_by_keys(BUFFER)).to(get_working_device()))
            elif n.type == ConstantHolder:
                self.add_module(n.name, node_builder(n))
                self.get_submodule(n.name).register_buffer(n.name,
                                                           torch.Tensor(n.get_weights_by_keys(CONSTANT)).to(get_working_device()))

            else:
                # Create a wrapper based on the corresponding quantization config.
                layer_wrapper = QuantizedLayerWrapper(n, get_quantization_config(n))
                # Add the wrapped layer to the model.
                self.add_module(n.name, layer_wrapper)

    def _get_op_func(self,
                     node: BaseNode,
                     configurable_nodes_names: List[str]) -> Any:
        """
        Get the operator corresponding to the passed node.

        Args:
            node: Node to get its op.
            configurable_nodes_names: List of nodes that are configurable.

        Returns:
            Operator (module) of the node.
        """
        return getattr(self, node.name)

    def _quantize_node_activations(self,
                                   node: BaseNode,
                                   input_tensors: List[torch.Tensor]) -> List[torch.Tensor]:
        """
        Quantize node's activation given input tensors.

        Args:
            node: Node to quantize its outputs.
            input_tensors: Input tensors of the node.

        Returns:
            Output of the node.

        """
        return input_tensors





class FullyQuantizedPyTorchModelBuilder(PyTorchModelBuilder):
    """
    Fully-Quantized PyTorch model.
    """

    def __init__(self,
                 graph: common.Graph):
        """

        Args:
            graph: Graph to build the model from.
        """

        super().__init__(graph)

    def build_model(self) -> Tuple[PytorchModel, UserInformation]:
        """
        Build a PyTorch fully quantized model and return it.
        Returns: Fully quantized PyTorch model and user information.

        """
        return FullyQuantizedPyTorchModel(self.graph), self.graph.user_info



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from typing import Callable

from model_compression_toolkit import CoreConfig
from model_compression_toolkit.core import common
from model_compression_toolkit.core.analyzer import analyzer_model_quantization
from model_compression_toolkit.core.common import Logger
from model_compression_toolkit.core.common.constants import TENSORFLOW, FOUND_TF
from model_compression_toolkit.core.common.mixed_precision.kpi_tools.kpi import KPI
from model_compression_toolkit.core.common.mixed_precision.mixed_precision_quantization_config import \
    MixedPrecisionQuantizationConfigV2
from model_compression_toolkit.core.common.target_platform.targetplatform2framework import TargetPlatformCapabilities
from model_compression_toolkit.core.exporter import export_model
from model_compression_toolkit.core.runner import core_runner, _init_tensorboard_writer
from model_compression_toolkit.ptq.runner import ptq_runner

if FOUND_TF:
    from model_compression_toolkit.core.keras.default_framework_info import DEFAULT_KERAS_INFO
    from model_compression_toolkit.core.keras.keras_implementation import KerasImplementation
    from model_compression_toolkit.core.keras.keras_model_validation import KerasModelValidation
    from tensorflow.keras.models import Model
    from model_compression_toolkit.core.keras.constants import DEFAULT_TP_MODEL
    from model_compression_toolkit.exporter import get_fully_quantized_keras_model

    from model_compression_toolkit import get_target_platform_capabilities
    DEFAULT_KERAS_TPC = get_target_platform_capabilities(TENSORFLOW, DEFAULT_TP_MODEL)


    def keras_post_training_quantization_experimental(in_model: Model,
                                                      representative_data_gen: Callable,
                                                      target_kpi: KPI = None,
                                                      core_config: CoreConfig = CoreConfig(),
                                                      target_platform_capabilities: TargetPlatformCapabilities = DEFAULT_KERAS_TPC,
                                                      new_experimental_exporter: bool = False):
        """
         Quantize a trained Keras model using post-training quantization. The model is quantized using a
         symmetric constraint quantization thresholds (power of two).
         The model is first optimized using several transformations (e.g. BatchNormalization folding to
         preceding layers). Then, using a given dataset, statistics (e.g. min/max, histogram, etc.) are
         being collected for each layer's output (and input, depends on the quantization configuration).
         For each possible bit width (per layer) a threshold is then being calculated using the collected
         statistics. Then, if given a mixed precision config in the core_config, using an ILP solver we find
         a mixed-precision configuration, and set a bit-width for each layer. The model is then quantized
         (both coefficients and activations by default).
         In order to limit the maximal model's size, a target KPI need to be passed after weights_memory
         is set (in bytes).

         Args:
             in_model (Model): Keras model to quantize.
             representative_data_gen (Callable): Dataset used for calibration.
             target_kpi (KPI): KPI object to limit the search of the mixed-precision configuration as desired.
             core_config (CoreConfig): Configuration object containing parameters of how the model should be quantized, including mixed precision parameters.
             target_platform_capabilities (TargetPlatformCapabilities): TargetPlatformCapabilities to optimize the Keras model according to.
             new_experimental_exporter (bool): Whether exporting the quantized model using new exporter or not (in progress. Avoiding it for now is recommended).

         Returns:

             A quantized model and information the user may need to handle the quantized model.

         Examples:

             Import MCT:

             >>> import model_compression_toolkit as mct

             Import a Keras model:

             >>> from tensorflow.keras.applications.mobilenet_v2 import MobileNetV2
             >>> model = MobileNetV2()

             Create a random dataset generator:

             >>> import numpy as np
             >>> def repr_datagen(): return [np.random.random((1,224,224,3))]

             Create a MCT core config, containing the quantization configuration and set the number of calibration iterations to 1:

             >>> config = mct.CoreConfig(n_iter=1)

             If mixed precision is desired, create a MCT core config with a mixed-precision configuration, to quantize a model with different bitwidths for different layers.
             The candidates bitwidth for quantization should be defined in the target platform model.
             In this example we use 1 image to search mixed-precision configuration:

             >>> config = mct.CoreConfig(n_iter=1, mixed_precision_config=mct.MixedPrecisionQuantizationConfigV2(num_of_images=1))

             For mixed-precision set a target KPI object:
             Create a KPI object to limit our returned model's size. Note that this value affects only coefficients
             that should be quantized (for example, the kernel of Conv2D in Keras will be affected by this value,
             while the bias will not):

             >>> kpi = mct.KPI(model.count_params() * 0.75)  # About 0.75 of the model size when quantized with 8 bits.

             Pass the model, the representative dataset generator, the configuration and the target KPI to get a
             quantized model:

             >>> quantized_model, quantization_info = mct.keras_post_training_quantization_experimental(model, repr_datagen, kpi, core_config=config)

             For more configuration options, please take a look at our `API documentation <https://sony.github.io/model_optimization/api/api_docs/modules/mixed_precision_quantization_config.html>`_.

         """

        fw_info = DEFAULT_KERAS_INFO

        KerasModelValidation(model=in_model,
                             fw_info=fw_info).validate()

        if core_config.mixed_precision_enable:
            if not isinstance(core_config.mixed_precision_config, MixedPrecisionQuantizationConfigV2):
                common.Logger.error("Given quantization config to mixed-precision facade is not of type "
                                    "MixedPrecisionQuantizationConfigV2. Please use keras_post_training_quantization API,"
                                    "or pass a valid mixed precision configuration.")

            common.Logger.info("Using experimental mixed-precision quantization. "
                               "If you encounter an issue please file a bug.")

        tb_w = _init_tensorboard_writer(fw_info)

        fw_impl = KerasImplementation()

        tg, bit_widths_config = core_runner(in_model=in_model,
                                            representative_data_gen=representative_data_gen,
                                            core_config=core_config,
                                            fw_info=fw_info,
                                            fw_impl=fw_impl,
                                            tpc=target_platform_capabilities,
                                            target_kpi=target_kpi,
                                            tb_w=tb_w)

        tg = ptq_runner(tg, representative_data_gen, core_config, fw_info, fw_impl, tb_w)

        if core_config.debug_config.analyze_similarity:
            analyzer_model_quantization(representative_data_gen,
                                        tb_w, tg,
                                        fw_impl,
                                        fw_info)

        if new_experimental_exporter:
            Logger.warning('Using new experimental exported models. '
                           'Please do not use unless you are familiar with what you are doing')
            return get_fully_quantized_keras_model(tg)

        return export_model(tg,
                            fw_info,
                            fw_impl,
                            tb_w,
                            bit_widths_config)



else:
    # If tensorflow or tensorflow_model_optimization are not installed,
    # we raise an exception when trying to use these functions.
    def keras_post_training_quantization_experimental(*args, **kwargs):
        Logger.critical('Installing tensorflow and tensorflow_model_optimization is mandatory '
                        'when using keras_post_training_quantization_experimental. '
                        'Could not find Tensorflow package.')



from tensorflow.keras.models import Model

from model_compression_toolkit import FrameworkInfo
from model_compression_toolkit.core.common.framework_info import ChannelAxis
from model_compression_toolkit.core.common.model_validation import ModelValidation
from model_compression_toolkit.core.keras.constants import CHANNELS_FORMAT, CHANNELS_FORMAT_LAST, CHANNELS_FORMAT_FIRST


class KerasModelValidation(ModelValidation):
    """
    Class to define validation methods in order to validate the received Keras model to quantize.
    """

    def __init__(self, model: Model, fw_info: FrameworkInfo):
        """
        Initialize a KerasModelValidation object.

        Args:
            model: Keras model to check its validity.
            fw_info: Information about the framework of the model (Keras).
        """

        super(KerasModelValidation, self).__init__(model=model,
                                                   fw_info=fw_info)

    def validate_output_channel_consistency(self):
        """

        Validate that output channels index in all layers of the model are the same.
        If the model has layers with different output channels index, an exception is thrown.

        """
        for layer in self.model.layers:
            data_format = layer.get_config().get(CHANNELS_FORMAT)
            if data_format is not None:
                assert (data_format == CHANNELS_FORMAT_LAST and self.fw_info.out_channel_axis_mapping.get(layer) == ChannelAxis.NHWC.value
                        or data_format == CHANNELS_FORMAT_FIRST and self.fw_info.out_channel_axis_mapping.get(layer) == ChannelAxis.NCHW.value), \
                    f'Model can not have layers with different data formats.'
